#include "ground_segmentation_alg.h"
#include "tgs.hpp"
#include "aos.hpp"

GroundSegmentationAlgorithm::GroundSegmentationAlgorithm(void)
{
  performance_supervisor_ptr_ = boost::make_shared<CTimeSupervisor>();
  pthread_mutex_init(&this->access_, NULL);
  std_dev_multiplier_ = 3.0;
}

GroundSegmentationAlgorithm::~GroundSegmentationAlgorithm(void)
{
  pthread_mutex_destroy(&this->access_);
}

void GroundSegmentationAlgorithm::config_update(Config &config, uint32_t level)
{
  this->lock();

  // save the current configuration
  this->config_ = config;

  this->unlock();
}

// TwoSegmentedInputsVehicleDetectorAlgorithm Public API

void GroundSegmentationAlgorithm::showPerformanceStatistics(void)
{
  performance_supervisor_ptr_->printStatistics();
}

void GroundSegmentationAlgorithm::segmentGround(
    const std_msgs::Header common_header, pcl::PointCloud<pcl::PointXYZRGBNormal> &lidar_pcl_cloud,
    pcl::PointCloud<pcl::PointXYZRGBNormal> &roadmap_vertices_pcl_cloud,
    iri_ground_segmentation::edges_array &edges_array,
    pcl::PointCloud<pcl::PointXYZRGBNormal> &ground_dense_reconstruction_pcl_cloud,
    visualization_msgs::MarkerArray &roadmap_markers)
{
  CFunctionMonitor performance_monitor("segmentGround", performance_supervisor_ptr_);
  std::vector<std::vector<int>> edges(100000);
  std::vector < std::vector < kf_based_terrain_analysis_lib::Edge >> discarded_edges(100000);

  bool use_travel = false;
  if (use_travel)
  {
    // Using TRAVEL for comparison!
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_in;
    pcl::PointCloud<pcl::PointXYZI>::Ptr ground_pc;
    pcl::PointCloud<pcl::PointXYZI>::Ptr nonground_pc;

    cloud_in.reset(new pcl::PointCloud<pcl::PointXYZI>());
    ground_pc.reset(new pcl::PointCloud<pcl::PointXYZI>());
    nonground_pc.reset(new pcl::PointCloud<pcl::PointXYZI>());

    boost::shared_ptr < travel::TravelGroundSeg < pcl::PointXYZI >> travel_ground_seg;

    travel_ground_seg.reset(new travel::TravelGroundSeg<pcl::PointXYZI>());

    double min_range = 1.0;
    double max_range = 64.0;

    double resolution = 8.0;
    int num_iter = 3;
    int num_lpr = 5;
    int num_min_pts = 10;
    double th_seeds = 0.5;
    double th_dist = 0.125;
    double th_outlier = 0.3;
    double th_normal = 0.940;
    double th_weight = 200;
    double th_lcc_normal = 0.03;
    double th_lcc_planar = 0.1;
    double th_obstacle = 1.0;
    bool refine_mode = true;
    bool visualization = true;

    std::cout << "Setting TRAVEL parameters!" << std::endl;
    travel_ground_seg->setParams(max_range, min_range, resolution, num_iter, num_lpr, num_min_pts, th_seeds, th_dist,
                                 th_outlier, th_normal, th_weight, th_lcc_normal, th_lcc_planar, th_obstacle,
                                 refine_mode, visualization);
    std::cout << "Parameters set!" << std::endl;

    std::cout << "Copying input data to XYZI format!" << std::endl;
    for (pcl::PointCloud<pcl::PointXYZRGBNormal>::iterator i = lidar_pcl_cloud.begin(); i != lidar_pcl_cloud.end(); ++i)
    {
      pcl::PointXYZI point;
      point.x = i->x;
      point.y = i->y;
      point.z = i->z;
      point.intensity = 0;
      cloud_in->points.push_back(point);
    }
    std::cout << "Data copied!" << std::endl;

    double ground_seg_time = 0.0;
    std::cout << "Starting TRAVEL ground segmentation!" << std::endl;
    travel_ground_seg->estimateGround(*cloud_in, *ground_pc, *nonground_pc, ground_seg_time);
    std::cout << "\033[1;35m Traversable-Ground Seg: " << cloud_in->size() << " -> Ground: " << ground_pc->size()
        << ", NonGround: " << nonground_pc->size() << "\033[0m" << std::endl;
    std::cout << "Traversable-Ground Seg time: " << ground_seg_time << std::endl;

    pcl::KdTreeFLANN < pcl::PointXYZI > kdtree_ground;
    kdtree_ground.setInputCloud(ground_pc);

    int K = 1;
    std::vector<int> ground_pointIdxKNNSearch(K);
    std::vector<float> ground_pointKNNSquaredDistance(K);

    int index = 0;
    for (pcl::PointCloud<pcl::PointXYZI>::iterator i = cloud_in->begin(); i != cloud_in->end(); ++i)
    {
      kdtree_ground.nearestKSearch(*i, K, ground_pointIdxKNNSearch, ground_pointKNNSquaredDistance);

      bool ground = false;
      if (sqrt(ground_pointKNNSquaredDistance[0]) < 0.001)
        ground = true;

      if (ground)
      {
        lidar_pcl_cloud.points[index].r = 0;
        lidar_pcl_cloud.points[index].g = 0;
        lidar_pcl_cloud.points[index].b = 255;

        lidar_pcl_cloud.points[index].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_GROUND + 0.995;
      }
      else
      {
        lidar_pcl_cloud.points[index].r = 0;
        lidar_pcl_cloud.points[index].g = 255;
        lidar_pcl_cloud.points[index].b = 0;

        lidar_pcl_cloud.points[index].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_OBSTACLE + 0.0;
      }

      ++index;
    }
  }
  else
  {
    //std::cout << "Calling to processor_.groundSegmentation!" << std::endl;

    processor_.groundSegmentation(lidar_configuration_, filtering_configuration_, lidar_pcl_cloud,
                                  roadmap_vertices_pcl_cloud, edges, discarded_edges,
                                  ground_dense_reconstruction_pcl_cloud);

    //std::cout << "Number of vertices = " << roadmap_vertices_pcl_cloud.points.size();
    //std::cout << "Number of edge vectors = " << edges.size() << std::endl;

    // Now we create the markers for visualization and change the format of edges to implement the interface
    // between the library and the node

    int vertex_id = 0;
    int discarded_vertex_id = 0;
    int edge_id = 0;
    int discarded_edge_id = 0;

    for (pcl::PointCloud<pcl::PointXYZRGBNormal>::iterator i = roadmap_vertices_pcl_cloud.begin();
        i != roadmap_vertices_pcl_cloud.end(); ++i)
    {
      pcl::PointXYZRGBNormal ground_ref = *i;

      // DEBUG!
      //std::cout << "Vertex Pitch = " << ground_ref.data_c[GRND_REF_DATA_C_2_PITCH] << " ;     Vertex Roll = "
      //          << ground_ref.data_c[GRND_REF_DATA_C_1_ROLL] << std::endl;

      //assert(
      //    vertex_id == ground_ref.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]
       //       && "Something unexpected in GroundSegmentationAlgorithm::segmentGround happened! vertex_id not equal to ground_ref.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]!");

      //std::cout << "Converting vertex with id = " << vertex_id << " to marker!" << std::endl;
      //std::cout << "Number of edges for vertex " << vertex_id << " = " << edges[ground_ref.data_n[3]].size() << std::endl;

      //  TODO: comment out the marker array creation to see how much time it consumes
      visualization_msgs::Marker vertex;
      vertex.action = visualization_msgs::Marker::ADD;
      vertex.header = common_header;

      if (ground_ref.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] < VERTEX_NOT_CONNECTED_TO_ROOT)
      {
        vertex.ns = "vertices_connected_to_root";
        vertex.id = vertex_id++;
      }
      else
      {
        vertex.ns = "vertices_NOT_connected_to_root";
        vertex.id = discarded_vertex_id++;
      }

      vertex.type = visualization_msgs::Marker::SPHERE;

      vertex.pose.position.x = ground_ref.x;
      vertex.pose.position.y = ground_ref.y;
      vertex.pose.position.z = ground_ref.z;

      vertex.pose.orientation.x = 0.0;
      vertex.pose.orientation.y = 0.0;
      vertex.pose.orientation.z = 0.0;
      vertex.pose.orientation.w = 1.0;

      vertex.scale.x = 0.25;
      vertex.scale.y = 0.25;
      vertex.scale.z = 0.25;

      vertex.color.a = 1.0; // Don't forget to set the alpha!
      if (ground_ref.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] == VERTEX_CONNECTED_TO_ROOT)
      {
        vertex.color.r = 1.0;
        vertex.color.g = 0.66;
        vertex.color.b = 0.0;
      }
      else
      {
        vertex.color.r = 1.0;
        vertex.color.g = 0.0;
        vertex.color.b = 0.0;
      }
      vertex.lifetime = ros::Duration(0); //0.1

      roadmap_markers.markers.push_back(vertex);

      iri_ground_segmentation::edges edges_temp;
      for (std::vector<int>::iterator e = edges[ground_ref.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].begin();
          e != edges[ground_ref.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].end(); e++)
      {
        edges_temp.edges.push_back(*e);

        pcl::PointXYZRGBNormal child_ground_ref = roadmap_vertices_pcl_cloud.points[*e];

        visualization_msgs::Marker edge;
        edge.action = visualization_msgs::Marker::ADD;
        edge.header = common_header;

        if (ground_ref.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] == VERTEX_CONNECTED_TO_ROOT &&
            child_ground_ref.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] == VERTEX_CONNECTED_TO_ROOT)
        {
          edge.ns = "edges";
          edge.id = edge_id++;
        }
        else
        {
          edge.ns = "discarded_edges";
          edge.id = discarded_edge_id++;
        }

        edge.type = visualization_msgs::Marker::LINE_LIST;

        geometry_msgs::Point p;
        p.x = ground_ref.x;
        p.y = ground_ref.y;
        p.z = ground_ref.z;
        edge.points.push_back(p);

        p.x = child_ground_ref.x;
        p.y = child_ground_ref.y;
        p.z = child_ground_ref.z;
        edge.points.push_back(p);

        edge.scale.x = 0.1;

        edge.pose.orientation.x = 0.0;
        edge.pose.orientation.y = 0.0;
        edge.pose.orientation.z = 0.0;
        edge.pose.orientation.w = 1.0;

        edge.color.a = 1.0; // Don't forget to set the alpha!
        if (ground_ref.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] == VERTEX_CONNECTED_TO_ROOT &&
            child_ground_ref.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] == VERTEX_CONNECTED_TO_ROOT)
        {
          edge.color.r = 0.8;
          edge.color.g = 0.53;
          edge.color.b = 0.0;
        }
        else
        {
          edge.color.r = R_VERTEX_NOT_CONNECTED_TO_ROOT / 255.0;
          edge.color.g = G_VERTEX_NOT_CONNECTED_TO_ROOT / 255.0;
          edge.color.b = B_VERTEX_NOT_CONNECTED_TO_ROOT / 255.0;
        }
        edge.lifetime = ros::Duration(0); // 0.1

        roadmap_markers.markers.push_back(edge);
      }

      edges_array.edges_array.push_back(edges_temp);

      // We create markers for the discarded edges
      for (int e = 0; e < discarded_edges[ground_ref.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].size(); e++)
      {
        kf_based_terrain_analysis_lib::Edge discarded_edge_index_and_class =
            discarded_edges[ground_ref.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]][e];

        visualization_msgs::Marker discarded_edge;
        discarded_edge.action = visualization_msgs::Marker::ADD;
        discarded_edge.header = common_header;
        discarded_edge.ns = "discarded_edges";
        discarded_edge.id = discarded_edge_id++;
        discarded_edge.type = visualization_msgs::Marker::LINE_LIST;

        geometry_msgs::Point p;
        p.x = ground_ref.x;
        p.y = ground_ref.y;
        p.z = ground_ref.z;
        discarded_edge.points.push_back(p);

        pcl::PointXYZRGBNormal child_ground_ref =
            roadmap_vertices_pcl_cloud.points[discarded_edge_index_and_class.vertex_index];
        p.x = child_ground_ref.x;
        p.y = child_ground_ref.y;
        p.z = child_ground_ref.z;
        discarded_edge.points.push_back(p);

        discarded_edge.scale.x = 0.1;

        discarded_edge.pose.orientation.x = 0.0;
        discarded_edge.pose.orientation.y = 0.0;
        discarded_edge.pose.orientation.z = 0.0;
        discarded_edge.pose.orientation.w = 1.0;

        discarded_edge.color.a = 1.0;
        switch (discarded_edge_index_and_class.edge_class)
        {
          case CLASS_EDGE_DISCARDED_TOO_MUCH_MAHALANOBIS_DISTANCE:
            //std::cout << "Coloring edge too much Mahalanobis distance! class = "
            //    << discarded_edge_index_and_class.edge_class << std::endl;
            discarded_edge.color.r = (float)R_EDGE_DISCARDED_TOO_MUCH_MAHALANOBIS_DISTANCE / 255.0;
            discarded_edge.color.g = (float)G_EDGE_DISCARDED_TOO_MUCH_MAHALANOBIS_DISTANCE / 255.0;
            discarded_edge.color.b = (float)B_EDGE_DISCARDED_TOO_MUCH_MAHALANOBIS_DISTANCE / 255.0;
            break;
          default:
            //std::cout << "WARNING! Discarded edge class not known!! class = " << discarded_edge_index_and_class.edge_class
            //    << std::endl;
            discarded_edge.color.r = 1.0;
            discarded_edge.color.g = 1.0;
            discarded_edge.color.b = 1.0;
            break;
        }
        discarded_edge.lifetime = ros::Duration(0); // 0.1
        roadmap_markers.markers.push_back(discarded_edge);
      }
    }

    //std::cout << "Total number of edges = " << edge_id << std::endl;
    //getchar();
  }
}
